import time

from keycloak import KeycloakAuthenticationError, KeycloakOpenID, KeycloakOperationError
from loguru import logger


class Session(object):
    def __init__(self, host: str, username: str, password: str):
        self.keycloak_openid = KeycloakOpenID(
            server_url=host, client_id="orkg-pypi", realm_name="orkg"
        )
        self.username = username
        self.password = password
        self._login(time.time())

    def get_access_token(self) -> str:
        timestamp = time.time()
        if self.timestamp + self.jwt.get("expires_in") < timestamp:
            if self.timestamp + self.jwt.get("refresh_expires_in") < timestamp:
                logger.debug("Access token expired. Signing in using credentials...")
                self._login(timestamp)
            else:
                logger.debug(
                    "Access token expired. Using refresh token to refresh access token."
                )
                self._refresh_token(timestamp)
        return self.jwt.get("access_token")

    def _login(self, timestamp: float):
        self.jwt = self.keycloak_openid.token(
            grant_type="password", username=self.username, password=self.password
        )
        self.timestamp = timestamp

    def _refresh_token(self, timestamp: float):
        try:
            refresh_token = self.jwt.get("refresh_token")
            self.jwt = self.keycloak_openid.refresh_token(refresh_token)
            self.timestamp = timestamp
        except (KeycloakAuthenticationError, KeycloakOperationError):
            logger.debug("Access token refresh failed. Signing in using credentials...")
            self._login(timestamp)



import re
from enum import Enum
from typing import Optional


class OID(object):
    """
    Represents an ORKG ID
    """

    id: str
    type: Optional[str]

    def __init__(self, id: str, type: Optional[str] = None):
        """
        :param id: the id
        :param type: the type of the id (optional)
        """
        self.id = id
        self.type = type

    def __str__(self) -> str:
        return self.id

    def __repr__(self) -> str:
        return self.id

    def __eq__(self, other: "OID") -> bool:
        if not isinstance(other, OID):
            return False
        return self.id == other.id and self.type == other.type

    def __hash__(self) -> int:
        return hash(self.id)

    def __bool__(self) -> bool:
        return bool(self.id)

    def is_literal_id(self) -> bool:
        return re.match(r"^L\d+$", self.id) is not None

    def is_resource_id(self) -> bool:
        return re.match(r"^R\d+$", self.id) is not None

    def is_predicate_id(self) -> bool:
        return re.match(r"^P\d+$", self.id) is not None

    def is_class_id(self) -> bool:
        return re.match(r"^C\d+$", self.id) is not None

    def is_custom_id(self) -> bool:
        return (
            not self.is_literal_id()
            and not self.is_resource_id()
            and not self.is_predicate_id()
            and not self.is_class_id()
        )

    def to_orkg_json(self) -> dict:
        obj = {"@id": self.id}
        if self.type:
            obj["@type"] = self.type
        return obj


class Hosts(Enum):
    PRODUCTION = "https://orkg.org"
    SANDBOX = "https://sandbox.orkg.org"
    INCUBATING = "https://incubating.orkg.org"


class ComparisonType(Enum):
    MERGE = "MERGE"
    PATH = "PATH"


class ExportFormat(Enum):
    CSV = "CSV"
    HTML = "HTML"
    DATAFRAME = "DATAFRAME"
    XML = "XML"
    UNKNOWN = "UNKNOWN"


class ThingType(Enum):
    COMPARISON = "COMPARISON"
    DIAGRAM = "DIAGRAM"
    VISUALIZATION = "VISUALIZATION"
    DRAFT_COMPARISON = "DRAFT_COMPARISON"
    LIST = "LIST"
    REVIEW = "REVIEW"
    QUALITY_REVIEW = "QUALITY_REVIEW"
    PAPER_VERSION = "PAPER_VERSION"
    ANY = "ANY"
    UNKNOWN = "UNKNOWN"



from typing import Optional

from orkg.out import OrkgResponse
from orkg.utils import NamespacedClient


class SnapshotsClient(NamespacedClient):
    def single_of_resource(self, resource_id: str, snapshot_id: str) -> OrkgResponse:
        """
        Get a specific snapshot of a resource.

        :param resource_id: The ID of the resource.
        :param snapshot_id: The ID of the snapshot.
        :return: An OrkgResponse containing the snapshot data.
        """
        self.client.backend._append_slash = False
        response = (
            self.client.backend.resources(resource_id)
            .snapshots(snapshot_id)
            .GET(
                headers={
                    "Accept": "application/json",
                    "Content-Type": "application/json;charset=UTF-8",
                }
            )
        )
        return self.client.wrap_response(response)

    def all_of_resource(
        self, resource_id: str, template_id: Optional[str] = None
    ) -> OrkgResponse:
        """
        Get all snapshots of a specific resource.

        :param resource_id: The ID of the resource.
        :param template_id: The id of the template that was used to create the resource snapshot. (optional)
        :return: An OrkgResponse containing the list of snapshots (paged).
        """
        self.client.backend._append_slash = True
        backend_func = self.client.backend.resources(resource_id).snapshots.GET
        if template_id is not None:
            response = backend_func(params={"template_id": template_id})
        else:
            response = backend_func()
        return self.client.wrap_response(response)

    def create_one(
        self, resource_id: str, template_id: str, register_handle: bool = True
    ) -> OrkgResponse:
        """
        Create a new snapshot for a specific resource.

        :param resource_id: The ID of the resource.
        :param template_id: The id of the template that will be used for subgraph exploration when creating the snapshot.
        :param register_handle: Whether to register a persistent Handle identifier. (optional, default: true)
        :return: An OrkgResponse containing the created snapshot data.
        """
        self.client.backend._append_slash = False
        response = self.client.backend.resources(resource_id).snapshots.POST(
            json={"template_id": template_id, "register_handle": register_handle},
            headers={
                **self.auth,
                "Accept": "application/json",
                "Content-Type": "application/json;charset=UTF-8",
            },
        )
        if response.ok:
            return self.client.expand_response(response)
        return self.client.wrap_response(response)



import logging

from orkg import errors
from orkg.client import ORKG
from orkg.common import OID, ComparisonType, ExportFormat, Hosts, ThingType
from orkg.graph import subgraph
from orkg.version import __version__

try:  # Python 2.7+
    from logging import NullHandler
except ImportError:

    class NullHandler(logging.Handler):
        def emit(self, record):
            pass


logger = logging.getLogger("ORKG")
logger.addHandler(logging.NullHandler())



import json
from typing import Dict, List, Optional, Set

from tqdm import tqdm

from orkg.client.templates.codegen import OTFFunctionWriter
from orkg.client.templates.components import Template
from orkg.utils import NamespacedClient


class TemplatesClient(NamespacedClient):
    materialized_templates: Set[str] = set()

    def materialize_template(self, template_id: str):
        """
        Materialize a singular ORKG template as python function
        :param template_id: the template id
        :return: True if everything is OK
        """
        materialization_result = self._fetch_and_build_template_function(template_id)
        Template.wip_templates.clear()
        return materialization_result

    def materialize_templates(
        self, templates: Optional[List[str]] = None, verbose: bool = True
    ):
        """
        Materialize a list of templates (or all templates if not provided)
        :param templates: optional list of templates
        :param verbose: sets if the process shows a progress bar or suppresses it
        """
        if templates is None:
            templates = [
                template["id"]
                for template in self.client.resources.get(
                    params={"include": ["NodeShape"], "size": 1000}
                ).content
            ]
        iterator = (
            tqdm(templates, desc="Materializing templates") if verbose else templates
        )

        failed_ids = []
        for template_id in iterator:
            try:
                self.materialize_template(template_id)
            except Exception as e:
                error_str = str(e)
                if "list index out of range" in error_str:
                    failed_ids.append(
                        {"id": template_id, "error": "no template with that ID"}
                    )
                else:
                    failed_ids.append({"id": template_id, "error": error_str})
        if len(failed_ids) > 0:
            print(f"Failed to materialize the following templates: {failed_ids}")

    def _fetch_and_build_template_function(self, template_id: str):
        """
        Internal function to create python digital twins of ORKG templates
        :param template_id: template ID to build
        :return: True if everything is OK
        """
        OTFFunctionWriter.implement_function(self, template_id)
        return True

    def list_templates(self):
        """
        List the set of materialized template functions
        :return: list of strings
        """
        return list(self.materialized_templates)

    def get_template_specifications(self, template_id: str) -> str:
        """
        Return JSON representation of a template's specification (schema)
        :param template_id: the template to lookup
        :return: string representation of a JSON object
        """
        template = Template(self.client, template_id)
        result = {
            comp.get_clean_name(): f"A value of type ({comp.value_class_label})"
            for comp in template.components
        }
        return json.dumps(result, indent=4, sort_keys=True, default=str)

    def create_template_instance(self, template_id: str, instance: Dict) -> str:
        """
        Creates an instance of a given template by filling in
        the specifications of the template with the provided values

        :param template_id: the string representing the template ID
        :param instance: the json object that contains the components of the template instance
        """
        template = Template(self.client, template_id)
        obj = template.create_api_object()
        for key, value in instance.items():
            obj = obj.replace(key, f'"{value}"')
        json_object = json.loads(obj)
        return self.client.objects.add(params=json_object).content["id"]



from typing import Any, Dict, List, Literal, Optional

from orkg.out import OrkgResponse
from orkg.utils import NamespacedClient, dict_to_url_params, query_params


class ComparisonsClient(NamespacedClient):
    def create(
        self,
        title: str,
        description: str,
        contributions: List[str],
        research_fields: List[str],
        config: Dict[str, Any],
        data: Dict[str, Any],
        sdgs: Optional[List[str]] = None,
        visualizations: Optional[List[str]] = None,
        references: Optional[List[str]] = None,
        organizations: Optional[List[str]] = None,
        observatories: Optional[List[str]] = None,
        is_anonymized: bool = False,
        extraction_method: Literal["UNKNOWN", "MANUAL", "AUTOMATIC"] = "UNKNOWN",
        authors: Optional[List[Dict[str, Any]]] = None,
    ) -> OrkgResponse:
        """
        Create a new comparison with the given parameters.

        :param title: The title of the comparison
        :param description: The description of the comparison
        :param contributions: The ids of the contributions the comparison compares
        :param research_fields: The list of research fields the comparison will be assigned to
        :param config: The configuration of the comparison (contains: predicates, contributions, transpose, type)
        :param data: The data contained in the comparison
        :param sdgs: The set of ids of sustainable development goals (optional)
        :param visualizations: The list of IDs of visualizations assigned to the comparison (optional)
        :param references: The references to external sources that the comparison refers to (optional)
        :param organizations: The list of IDs of the organizations (optional)
        :param observatories: The list of IDs of the observatories (optional)
        :param is_anonymized: Whether the comparison should be displayed as anonymous (default: False)
        :param extraction_method: The method used to extract the comparison resource (default: UNKNOWN)
        :param authors: The list of authors that contributed to the comparison (optional)
        :return: The response of the request
        """
        payload = {
            "title": title,
            "description": description,
            "contributions": contributions,
            "research_fields": research_fields,
            "config": config,
            "data": data,
            "is_anonymized": is_anonymized,
            "extraction_method": extraction_method,
            "sdgs": sdgs if sdgs else [],
            "visualizations": visualizations if visualizations else [],
            "references": references if references else [],
            "organizations": organizations if organizations else [],
            "observatories": observatories if observatories else [],
            "authors": authors if authors else [],
        }

        self.client.backend._append_slash = False
        headers = {
            "Content-Type": "application/vnd.orkg.comparison.v2+json;charset=UTF-8",
            "Accept": "application/vnd.orkg.comparison.v2+json",
        }
        headers.update(self.auth)
        response = self.client.backend.comparisons.POST(json=payload, headers=headers)
        if response.ok:
            return self.client.expand_response(response)
        return self.client.wrap_response(response)

    def publish(
        self,
        comparison_id: str,
        subject: str,
        description: str,
        authors: List[Dict[str, Any]],
        assign_doi: bool = False,
    ) -> OrkgResponse:
        """
        Publish an existing comparison with the given parameters.

        :param comparison_id: The ID of the comparison to publish
        :param subject: The subject of the comparison
        :param description: The description of the comparison
        :param authors: The list of authors that contributed to the comparison
        :param assign_doi: Whether to assign a new DOI for the comparison (default: False)
        :return: The response of the request
        """
        payload = {
            "subject": subject,
            "description": description,
            "authors": authors,
            "assign_doi": assign_doi,
        }

        self.client.backend._append_slash = False
        headers = {"Content-Type": "application/json", "Accept": "application/json"}
        headers.update(self.auth)
        response = self.client.backend.comparisons(comparison_id).publish.POST(
            json=payload, headers=headers
        )
        if response.ok:
            return self.client.expand_response(response)
        return self.client.wrap_response(response)

    @query_params("page", "size", "sort", "desc")
    def in_research_field(
        self,
        research_field_id: str,
        include_subfields: Optional[bool] = False,
        params: Optional[Dict] = None,
    ) -> OrkgResponse:
        """
        Get all comparisons in a research field
        :param research_field_id: the id of the research field
        :param include_subfields: True/False whether to include comparisons from subfields, default is False (optional)
        :param page: the page number (optional)
        :param size: number of items per page (optional)
        :param sort: key to sort on (optional)
        :param desc: true/false to sort desc (optional)
        :return: OrkgResponse object
        """
        url = self.client.backend("research-fields")(research_field_id)
        if include_subfields:
            url = url.subfields.comparisons
        else:
            url = url.comparisons
        if len(params) > 0:
            self.client.backend._append_slash = False
            response = url.GET(dict_to_url_params(params))
        else:
            self.client.backend._append_slash = True
            response = url.GET()
        return self.client.wrap_response(response)



from typing import Dict, Optional

from orkg.out import OrkgResponse, OrkgUnpaginatedResponse
from orkg.utils import NamespacedClient, admin_functionality, query_params


class ResourcesClient(NamespacedClient):
    def by_id(self, id: str) -> OrkgResponse:
        """
        Lookup a resource by id
        :param id: the id of the resource to lookup
        :return: an OrkgResponse object containing the resource
        """
        self.client.backend._append_slash = True
        response = self.client.backend.resources(id).GET()
        return self.client.wrap_response(response)

    @query_params("q", "exact", "page", "size", "sort", "desc", "exclude", "include")
    def get(self, params: Optional[Dict] = None) -> OrkgResponse:
        """
        Fetch a list of resources, with the possibility to paginate the results and filter them out based on label
        :param q: search term of the label of the resource (optional)
        :param exact: whether to check for the exact search term or not (optional) -> bool
        :param page: the page number (optional)
        :param size: number of items per page (optional)
        :param sort: key to sort on (optional)
        :param desc: true/false to sort desc (optional)
        :param exclude: classes to be excluded in search (optional)
        :param include: classes to filter the resources by (optional)
        :return: an OrkgResponse object contains the list of resources
        """
        self.handle_sort_params(params)
        response = self.client.backend.resources.GET(params=params)
        return self.client.wrap_response(response)

    @query_params("q", "exact", "page", "size", "sort", "desc", "exclude", "include")
    def get_unpaginated(
        self, start_page: int = 0, end_page: int = -1, params: Optional[Dict] = None
    ) -> OrkgUnpaginatedResponse:
        """
        Fetch all resources returned by all pages between start_page and end_page.

        :param q: search term of the label of the resource (optional)
        :param exact: whether to check for the exact search term or not (optional) -> bool
        :param page: the page number (optional)
        :param size: number of items per page (optional)
        :param sort: key to sort on (optional)
        :param desc: true/false to sort desc (optional)
        :param exclude: classes to be excluded in search (optional)
        :param include: classes to filter the resources by (optional)

        :param start_page: page to start from. Defaults to 0 (optional)
        :param end_page: page to stop at. Defaults to -1 meaning non-stop (optional)
        :return: an OrkgUnpaginatedResponse object
        """
        return self._call_pageable(
            self.get, args={}, params=params, start_page=start_page, end_page=end_page
        )

    @query_params("id", "label", "classes", "extraction_method")
    def add(self, params: Optional[Dict] = None) -> OrkgResponse:
        """
        Create a new resource in the ORKG instance
        :param id: the specific id to add (optional)
        :param label: the label of the new resource (optional)
        :param classes: list of classes to assign the resource to (optional)
        :param extraction_method: the extraction method of the resource (optional) one of [MANUAL, AUTOMATIC, UNKNOWN]
        :return: an OrkgResponse object containing the newly created resource
        """
        if len(params) == 0:
            raise ValueError("at least label should be provided")
        else:
            self.client.backend._append_slash = True
            response = self.client.backend.resources.POST(
                json=params, headers=self.auth
            )
            if response.ok:
                return self.client.expand_response(response)
        return self.client.wrap_response(response)

    @query_params("id", "label", "classes", "extraction_method")
    def find_or_add(self, params: Optional[Dict] = None) -> OrkgResponse:
        """
        find or create a new resource in the ORKG instance
        :param id: the specific id to add (optional)
        :param label: the label of the new resource (optional)
        :param classes: list of classes to assign the resource to (optional)
        :param extraction_method: the extraction method of the resource (optional) one of [MANUAL, AUTOMATIC, UNKNOWN]
        :return: an OrkgResponse object containing the found or newly created resource
        """
        if len(params) == 0:
            raise ValueError("at least a label should be provided")
        else:
            if "id" in params:
                # check if a resource with this id is there
                found = self.by_id(params["id"])
                if found.succeeded:
                    return found
            if "label" in params and "classes" in params:
                class_id = params["classes"][0]
                # check if a resource exists with this label and the _first_ class from params class list
                found = self.client.resources.get(
                    params={
                        "include": [class_id],
                        "q": params["label"],
                        "exact": True,
                        "size": 1,
                    }
                )
                if found.succeeded:
                    if isinstance(found.content, list) and len(found.content) > 0:
                        found.content = found.content[0]
                        return found
            # if no class is passed, check if a resource with this label is there
            elif "label" in params:
                found = self.get(q=params["label"], exact=True, size=1)
                if found.succeeded:
                    if isinstance(found.content, list) and len(found.content) > 0:
                        found.content = found.content[0]
                        return found
            # None found, what the hell! let's create a new one
            self.client.backend._append_slash = True
            response = self.client.backend.resources.POST(
                json=params, headers=self.auth
            )
            if response.ok:
                return self.client.expand_response(response)
        return self.client.wrap_response(response)

    @query_params("label", "classes", "extraction_method")
    def update(self, id: str, params: Optional[Dict] = None) -> OrkgResponse:
        """
        Update a resource with a specific id
        :param id: the id of the resource to update
        :param label: the new label (optional)
        :param classes: the updated list of classes (optional)
        :param extraction_method: the extraction method of the resource (optional) one of [MANUAL, AUTOMATIC, UNKNOWN]
        :return: an OrkgResponse object contains the newly updated resource
        """
        if len(params) == 0:
            raise ValueError("at least label should be provided")
        else:
            if not self.exists(id):
                raise ValueError("the provided id is not in the graph")
            self.client.backend._append_slash = True
            response = self.client.backend.resources(id).PUT(
                json=params, headers=self.auth
            )
            if response.ok:
                return self.client.expand_response(response)
        return self.client.wrap_response(response)

    @admin_functionality
    def update_observatory(
        self, id: str, observatory_id: str, organization_id: str
    ) -> OrkgResponse:
        """
        Update the observatory and the organization of a particular resource
        NOTE: this is a call only for admins
        :param id: The id of the resource to update
        :param observatory_id: the new observatory id
        :param organization_id: the new organization id
        :return: an OrkgResponse object contains the newly updated resource
        """
        if not self.exists(id):
            raise ValueError("the provided id is not in the graph")
        self.client.backend._append_slash = False
        content = {"observatory_id": observatory_id, "organization_id": organization_id}
        response = self.client.backend.resources(id).PUT(
            json=content, headers=self.auth
        )
        if response.ok:
            return self.client.expand_response(response)
        return self.client.wrap_response(response)

    def exists(self, id: str) -> bool:
        """
        Check if a resource exists in the graph
        :param id: the id of the resource to check
        :return: true if found, otherwise false
        """
        return self.by_id(id).succeeded

    @admin_functionality
    def delete(self, id: str) -> OrkgResponse:
        if not self.exists(id):
            raise ValueError("the provided id is not in the graph")
        self.client.backend._append_slash = True
        response = self.client.backend.resources(id).DELETE(headers=self.auth)
        return self.client.wrap_response(response)



from typing import Optional, Union

import hammock
import requests
from requests import HTTPError
from requests.adapters import HTTPAdapter

from orkg.common import Hosts
from orkg.logging_config import initialize_logger
from orkg.out import OrkgResponse
from orkg.utils import check_authentication, follow_location_header, verb_logger
from orkg.version import __version__

from .classes import ClassesClient
from .comparisons import ComparisonsClient
from .contribution_comparisons import ContributionComparisonsClient
from .contributions import ContributionsClient
from .dummy import DummyClient
from .fields import ResearchFieldsClient
from .harvesters import HarvestersClient
from .json import JSONClient
from .lists import ListsClient
from .literals import LiteralsClient
from .objects import ObjectsClient
from .papers import PapersClient
from .predicates import PredicatesClient
from .problems import ResearchProblemsClient
from .resources import ResourcesClient
from .session import Session
from .snapshots import SnapshotsClient
from .statements import StatementsClient
from .stats import StatsClient
from .templates import TemplatesClient

hammock.Hammock.GET = verb_logger("GET")(hammock.Hammock.GET)
hammock.Hammock.POST = check_authentication(verb_logger("POST")(hammock.Hammock.POST))
hammock.Hammock.PUT = check_authentication(verb_logger("PUT")(hammock.Hammock.PUT))
hammock.Hammock.DELETE = check_authentication(
    verb_logger("DELETE")(hammock.Hammock.DELETE)
)
hammock.Hammock.PATCH = check_authentication(
    verb_logger("PATCH")(hammock.Hammock.PATCH)
)

DEFAULT_USER_AGENT = f"ORKG Python Client/{__version__}"


class TimeoutHTTPAdapter(HTTPAdapter):
    def __init__(self, timeout=None, *args, **kwargs):
        self.timeout = timeout
        super().__init__(*args, **kwargs)

    def send(self, request, **kwargs):
        kwargs["timeout"] = kwargs.get("timeout") or self.timeout
        return super().send(request, **kwargs)


class ORKG(object):
    """
    The base class for the ORKG client.
    Contains all the methods to interact with the ORKG API.
    """

    host: Optional[Union[str, Hosts]] = None
    simcomp_host: Optional[str] = None
    session: Optional[Session] = None

    def __init__(
        self,
        host: Optional[Union[str, Hosts]] = None,
        auth_host: Optional[str] = "https://accounts.orkg.org",
        simcomp_host: Optional[str] = None,
        creds: Optional[tuple] = None,
        logging_level: Union[str, int] = "WARNING",
        follow_location: bool = True,
        timeout: Optional[float] = None,
        user_agent: Optional[str] = None,
        **kwargs,
    ):
        self._set_host(host)
        self._set_simcomp_host(simcomp_host)

        # Set timeout for hammock sessions
        timeout_adapter = TimeoutHTTPAdapter(timeout=timeout)
        session = requests.Session()
        session.mount("http://", timeout_adapter)
        session.mount("https://", timeout_adapter)

        # Set User-Agent header
        self.user_agent = user_agent or DEFAULT_USER_AGENT
        session.headers["User-Agent"] = self.user_agent

        self.core = hammock.Hammock(self.host, **kwargs)
        self.core._session = session
        if self.simcomp_available:
            self.simcomp = hammock.Hammock(self.simcomp_host, **kwargs)
            self.simcomp._session = session

        initialize_logger(logging_level)
        self.follow_location = follow_location
        if creds is not None and len(creds) == 2:
            self.session = Session(auth_host, creds[0], creds[1])
        else:
            self.session = None
        self.backend = self.core.api
        self.resources = ResourcesClient(self)
        self.predicates = PredicatesClient(self)
        self.classes = ClassesClient(self)
        self.lists = ListsClient(self)
        self.literals = LiteralsClient(self)
        self.stats = StatsClient(self)
        self.statements = StatementsClient(self)
        self.papers = PapersClient(self)
        self.comparisons = ComparisonsClient(self)
        self.contributions = ContributionsClient(self)
        self.objects = ObjectsClient(self)
        self.dummy = DummyClient(self)
        self.json = JSONClient(self)
        self.templates = TemplatesClient(self)
        self.contribution_comparisons = ContributionComparisonsClient(self)
        self.harvesters = HarvestersClient(self)
        self.problems = ResearchProblemsClient(self)
        self.fields = ResearchFieldsClient(self)
        self.snapshots = SnapshotsClient(self)

    @property
    def simcomp_available(self) -> bool:
        """
        Check whether the simcomp host is available or not
        :return: True if the simcomp host is available, False otherwise
        """
        return self.simcomp_host is not None

    def _set_host(self, host: Optional[Union[str, Hosts]]) -> None:
        """
        Set the host value of the backend
        :param host: the host passed by the user or None
        """
        if isinstance(host, Hosts):
            host = host.value
        if host is not None and not host.startswith("http"):
            if "host" not in host.lower():
                raise ValueError("host must begin with http or https")
            else:
                raise ValueError(
                    "the host name was not recognized "
                    "-- use Hosts.PRODUCTION, Hosts.SANDBOX, or Hosts.INCUBATING without quotes"
                )
        if host is not None and host[-1] == "/":
            host = host[:-1]
        self.host = host if host is not None else "https://sandbox.orkg.org"

    def _set_simcomp_host(self, simcomp_host: Optional[str]) -> None:
        """
        Set the simcomp host
        :param simcomp_host: the simcomp host passed by the user (optional)
        """
        if simcomp_host is not None and not simcomp_host.startswith("http"):
            raise ValueError("simcomp host must begin with http or https")
        if simcomp_host is None and "orkg" in self.host:
            simcomp_host = self.host + "/simcomp"
        self.simcomp_host = simcomp_host

    def ping(self) -> bool:
        """
        The ping function checks whether the backend is live or not.
        :returns: True if the backend is live, False otherwise
        """
        try:
            initial_response = requests.get(self.host, timeout=3, allow_redirects=True)
            initial_response.raise_for_status()
            return True
        except HTTPError:
            # TODO: log different status errors
            return False

    def expand_response(self, response):
        url: str = response.headers.get("Location")
        if self.follow_location:
            original_headers = (
                dict(response.request.headers) if hasattr(response, "request") else None
            )
            if original_headers is not None:
                original_headers = {
                    k: v
                    for k, v in original_headers.items()
                    if k in ["Authorization", "Content-Type", "Accept", "User-Agent"]
                }
            return self.wrap_response(
                follow_location_header(url, headers=original_headers)
            )
        else:
            return self.wrap_response(
                status_code=str(response.status_code), content={}, url=url
            )

    def wrap_response(
        self,
        response=None,
        status_code: str = None,
        content: Union[list, dict, str] = None,
        url: str = None,
    ) -> OrkgResponse:
        """
        Wraps the response from the backend into a OrkgResponse object
        :param response: the response from the backend
        :param status_code: the status code of the response
        :param content: the content of the response
        :param url: the url of the response
        :return: OrkgResponse object
        """
        return OrkgResponse(
            client=self,
            response=response,
            status_code=status_code,
            content=content,
            url=url,
        )



from typing import Dict, Optional

from orkg.out import OrkgResponse
from orkg.utils import NamespacedClient, dict_to_url_params, query_params


class ListsClient(NamespacedClient):
    def by_id(self, list_id: str) -> OrkgResponse:
        """
        Lookup a list by ID to find the label and the IDs of its elements.
        :param list_id: the id of the list to lookup
        :return: an OrkgResponse object containing the list
        """
        is_list = "List" in self.client.resources.by_id(list_id).content["classes"]
        if not is_list:
            raise ValueError("the ID provided does not correspond to a list!")
        else:
            self.client.backend._append_slash = True
            response = self.client.backend.lists(list_id).GET()
        return self.client.wrap_response(response)

    @query_params("page", "size", "sort", "desc")
    def get_elements(self, list_id: str, params: Optional[Dict] = None) -> OrkgResponse:
        """
        Get the complete representations of the elements in a list.
        :param list_id: the id of the list to lookup
        :param page: the page number (optional)
        :param size: number of items per page (optional)
        :param sort: key to sort on (optional)
        :param desc: true/false to sort desc (optional)
        :return: an OrkgResponse object containing the elements in the list
        """
        is_list = "List" in self.client.resources.by_id(list_id).content["classes"]
        if not is_list:
            raise ValueError("the ID provided does not correspond to a list!")
        if len(params) > 0:
            self.client.backend._append_slash = False
            response = self.client.backend.lists(list_id).elements.GET(
                dict_to_url_params(params)
            )
        else:
            self.client.backend._append_slash = True
            response = self.client.backend.lists(list_id).elements.GET()
        return self.client.wrap_response(response)

    @query_params("label", "elements")
    def add(self, params: Optional[Dict] = None) -> OrkgResponse:
        """
        Create a new list with the given label and elements.
        :param label: string of the label of the list
        :param elements: list of string ids for the elements of the list
        :return: an OrkgResponse object containing the list
        """
        if len(params) != 2:
            raise ValueError("a label and a list of elements must be provided")
        else:
            self.client.backend._append_slash = True
            response = self.client.backend.lists.POST(json=params, headers=self.auth)
            if response.ok:
                return self.client.expand_response(response)
        return self.client.wrap_response(response)

    @query_params("label", "elements")
    def update(self, list_id: str, params: Optional[Dict] = None) -> OrkgResponse:
        """
        Replace the current label and/or elements of an existing list with new values.
        Does not return the modified list in the body of the response.
        :param list_id: the id of the list to lookup
        :param label: string of the new label of the list (optional)
        :param elements: list of string ids for the new elements of the list (optional)
        :return: an OrkgResponse object indicating success or failure
        """
        if len(params) == 0:
            raise ValueError("either a label or a list of elements must be provided")
        is_list = "List" in self.client.resources.by_id(list_id).content["classes"]
        if not is_list:
            raise ValueError("the ID provided does not correspond to a list!")
        else:
            self.client.backend._append_slash = True
            response = self.client.backend.lists(list_id).PATCH(
                json=params, headers=self.auth
            )
            if response.ok:
                return self.client.expand_response(response)
        return self.client.wrap_response(response)



import re
from ast import literal_eval
from typing import Dict, List, Optional, Tuple, Union

from requests import Response

from orkg.out import OrkgResponse, OrkgUnpaginatedResponse
from orkg.utils import NamespacedClient, dict_to_url_params, query_params


class PapersClient(NamespacedClient):
    def by_id(self, id: str) -> OrkgResponse:
        """
        Lookup a paper by id
        :param id: the id of the paper to lookup
        :return: an OrkgResponse object containing the paper
        """
        self.client.backend._append_slash = True
        response = self.client.backend.papers(id).GET()
        return self.client.wrap_response(response)

    @query_params(
        "title",
        "exact",
        "doi",
        "visibility",
        "verified",
        "created_by",
        "created_at_start",
        "created_at_end",
        "observatory_id",
        "organization_id",
        "research_field",
        "include_subfields",
        "page",
        "size",
        "sort",
        "desc",
    )
    def get(self, params: Optional[Dict] = None) -> OrkgResponse:
        """
        Fetch a list of papers, with multiple options for searching and filtering results
        :param title: search term that must be contained in the title of the paper (optional)
        :param exact: whether title matching is exact or fuzzy (optional, default: false)
        :param doi: filter for the DOI of the paper (optional)
        :param visibility: possible values are "ALL_LISTED", "UNLISTED", "FEATURED", "NON_FEATURED", "DELETED" (optional)
        :param verified: filter for the verified flag of the paper (optional)
        :param created_by: filter for the UUID of the user or service who created this paper (optional)
        :param created_at_start: the oldest timestamp a returned paper can have (optional)
        :param created_at_end: the most recent timestamp a returned paper can have (optional)
        :param observatory_id: filter for the UUID of the observatory that the paper belongs to (optional)
        :param organization_id: filter for the UUID of the organization that the resource belongs to (optional)
        :param research_field: filter for research field id (optional)
        :param include_subfields: whether subfields are included or not (optional, default: false)
        :param page: the page number (optional)
        :param size: number of items per page (optional)
        :param sort: key to sort on (optional)
        :param desc: true/false to sort desc (optional)
        :return: an OrkgResponse object containing a list of papers
        """
        if len(params) > 0:
            self.handle_sort_params(params)
            self.client.backend._append_slash = False
            response = self.client.backend.papers.GET(
                params=params,
                headers={
                    "Content-Type": "application/vnd.orkg.paper.v2+json;charset=UTF-8",
                    "Accept": "application/vnd.orkg.paper.v2+json",
                },
            )
        else:
            self.client.backend._append_slash = False
            response = self.client.backend.papers.GET()
        return self.client.wrap_response(response)

    @query_params(
        "title",
        "exact",
        "doi",
        "visibility",
        "verified",
        "created_by",
        "created_at_start",
        "created_at_end",
        "observatory_id",
        "organization_id",
        "research_field",
        "include_subfields",
        "page",
        "size",
        "sort",
        "desc",
    )
    def get_unpaginated(
        self, start_page: int = 0, end_page: int = -1, params: Optional[Dict] = None
    ) -> OrkgUnpaginatedResponse:
        """
        Fetch a list of papers between start and end page, with multiple options for searching and filtering results
        :param start_page: page to start from. Defaults to 0 (optional)
        :param end_page: page to stop at. Defaults to -1 meaning non-stop (optional)
        :param title: search term that must be contained in the title of the paper (optional)
        :param exact: whether title matching is exact or fuzzy (optional, default: false)
        :param doi: filter for the DOI of the paper (optional)
        :param visibility: possible values are "ALL_LISTED", "UNLISTED", "FEATURED", "NON_FEATURED", "DELETED" (optional)
        :param verified: filter for the verified flag of the paper (optional)
        :param created_by: filter for the UUID of the user or service who created this paper (optional)
        :param created_at_start: the oldest timestamp a returned paper can have (optional)
        :param created_at_end: the most recent timestamp a returned paper can have (optional)
        :param observatory_id: filter for the UUID of the observatory that the paper belongs to (optional)
        :param organization_id: filter for the UUID of the organization that the resource belongs to (optional)
        :param research_field: filter for research field id (optional)
        :param include_subfields: whether subfields are included or not (optional, default: false)
        :param page: the page number (optional)
        :param size: number of items per page (optional)
        :param sort: key to sort on (optional)
        :param desc: true/false to sort desc (optional)
        :return: an OrkgUnpaginatedResponse object containing a list of papers
        """
        return self._call_pageable(
            self.get, args={}, params=params, start_page=start_page, end_page=end_page
        )

    def update(
        self,
        paper_id: str,
        title: Optional[str] = None,
        research_fields: Optional[List] = None,
        identifiers: Optional[Dict] = None,
        publication_info: Optional[Dict] = None,
        authors: Optional[List[Dict]] = None,
        organizations: Optional[List] = None,
        observatories: Optional[List] = None,
        sdgs: Optional[set] = None,
    ) -> OrkgResponse:
        """
        Edit the metadata of an already created paper. The passed values will replace the existing values.
        :param paper_id: the resource ID of the paper
        :param title: the updated title (optional)
        :param research_fields: the updated research fields of the paper (optional)
        :param identifiers: the updated unique identifiers of the paper (optional)
        :param publication_info: the updated publication info of the paper (optional)
        :param authors: the updated list of authors' info of the paper (optional)
        :param organizations: the updated list of IDs of the organizations the paper belongs to (optional)
        :param observatories: the updated list of IDs of the observatories the paper belongs to (optional)
        :param sdgs: the updated set of IDs of UN sustainable development goals the paper will be assigned to (optional)
        :return: OrkgResponse indicating success or failure of the request (no content)
        """
        paper = {
            "title": title,
            "research_fields": research_fields,
            "identifiers": identifiers,
            "publication_info": publication_info,
            "authors": authors,
            "organizations": organizations,
            "observatories": observatories,
            "sdgs": sdgs,
        }
        if all(value is None for value in paper.values()):
            raise ValueError("at least one field to update should be passed!")
        headers = {
            **(self.auth if self.auth is not None else {}),
            "Content-Type": "application/vnd.orkg.paper.v2+json;charset=UTF-8",
            "Accept": "application/vnd.orkg.paper.v2+json",
        }
        self.client.backend._append_slash = False
        response = self.client.backend.papers(paper_id).PUT(json=paper, headers=headers)
        if response.ok:
            return self.client.expand_response(response)
        return self.client.wrap_response(response)

    def publish(
        self, paper_id: str, subject: str, description: str, authors: List[Dict]
    ) -> OrkgResponse:
        """
        Publish an existing paper by assigning a DOI and adding publication info.
        :param paper_id: the resource ID of the paper
        :param subject: the subject of the paper
        :param description: the description of the paper
        :param authors: the list of authors' info
        :return: OrkgResponse indicating success or failure of the request (no content)
        """
        paper = {"subject": subject, "description": description, "authors": authors}
        headers = {
            **(self.auth if self.auth is not None else {}),
            "Content-Type": "application/json;charset=UTF-8",
            "Accept": "application/json",
        }
        self.client.backend._append_slash = False
        response = self.client.backend.papers(paper_id).publish.POST(
            json=paper, headers=headers
        )
        if response.ok:
            return self.client.expand_response(response)
        return self.client.wrap_response(response)

    def add(
        self, params: Optional[Dict] = None, merge_if_exists: bool = False
    ) -> OrkgResponse:
        """
        Create a new paper in the ORKG instance. Automatically detects whether the provided data structure is compatible
        with the legacy or the new endpoint and calls the corresponding method.
        :param params: dict containing the metadata and content of the paper
        :param merge_if_exists: merge the papers if they exist in the graph and append contributions (optional)
        :return: an OrkgResponse containing either the new paper resource (v1) or the url to the new paper object (v2)
        """
        if "paper" in params.keys():
            required_params = ["title", "researchField"]
            missing_params = [
                p for p in required_params if p not in params["paper"].keys()
            ]
            if missing_params:
                raise TypeError(
                    f"Params is missing the following required key(s): {missing_params}"
                )
            return self.add_v1(params, merge_if_exists)
        else:
            required_params = [
                "authors",
                "contents",
                "extraction_method",
                "identifiers",
                "observatories",
                "organizations",
                "publication_info",
                "research_fields",
                "title",
            ]
            missing_params = [p for p in required_params if p not in params.keys()]
            if missing_params:
                raise TypeError(
                    f"Params is missing the following required key(s): {missing_params}"
                )
            return self.add_v2(**params)

    def add_v1(self, params: Dict, merge_if_exists: bool = False) -> OrkgResponse:
        """
        Create a new paper or add contributions to an existing paper in the ORKG instance using the legacy papers endpoint.
        :param params: paper Object
        :param merge_if_exists: merge the papers if they exist in the graph and append contributions
        :return: an OrkgResponse object containing the newly created paper resource
        """
        self.client.backend._append_slash = True
        response = self.client.backend.papers.POST(
            json=params, params={"mergeIfExists": merge_if_exists}, headers=self.auth
        )
        if response.ok:
            return self.client.expand_response(response)
        return self.client.wrap_response(response)

    # TODO: remove self.auth check after implementing package-wide solution, see: https://gitlab.com/TIBHannover/orkg/orkg-pypi/-/issues/76
    def add_v2(
        self,
        title: str,
        research_fields: List,
        identifiers: Dict,
        publication_info: Dict,
        authors: List[Dict],
        contents: Dict,
        organizations: List,
        observatories: List,
        extraction_method: str,
        sdgs: Optional[set] = None,
        **kwargs,
    ) -> OrkgResponse:
        """
        Create a new paper in the ORKG instance using the new papers endpoint.
        :param title: the title of the paper
        :param research_fields: the list of research field IDs the paper will be assigned to
        :param identifiers: the unique identifiers of the paper (e.g. doi)
        :param publication_info: the publication info of the paper (e.g. year, url)
        :param authors: the authors of the paper
        :param contents: the contributions of the paper and resources, literals, predicates, and lists to be created
        :param organizations: the list of IDs of the organizations the paper belongs to
        :param observatories: the list of IDs of the observatories the paper belongs to
        :param extraction_method: the method used to extract the paper resource ("UNKNOWN", "MANUAL" or "AUTOMATIC")
        :param sdgs: the set of IDs of UN sustainable development goals the paper will be assigned to (optional)
        :return: an OrkgResponse containing the status code and url of the new paper object (no content)
        """
        assert len(kwargs) == 0, f"Unknown arguments: {list(kwargs.keys())}"
        paper = {
            "title": title,
            "research_fields": research_fields,
            "identifiers": identifiers,
            "publication_info": publication_info,
            "authors": authors,
            "contents": contents,
            "organizations": organizations,
            "observatories": observatories,
            "extraction_method": extraction_method,
            "sdgs": sdgs,
        }
        headers = {
            **(self.auth if self.auth is not None else {}),
            "Content-Type": "application/vnd.orkg.paper.v2+json;charset=UTF-8",
            "Accept": "application/vnd.orkg.paper.v2+json",
        }
        self.client.backend._append_slash = False
        response: Response = self.client.backend.papers.POST(
            json=paper, headers=headers
        )
        if response.ok:
            return self.client.expand_response(response)
        return self.client.wrap_response(response)

    def by_doi(self, doi: str) -> OrkgResponse:
        """
        Get the paper resource in the ORKG associated with the given DOI.
        :param doi: The DOI of the paper, could be a full url or just the DOI
        :return: A list of paper objects
        """
        # Disable appending a slash to the base URL
        self.client.backend._append_slash = False

        # Extract the DOI from the input if it's a full URL
        if "doi.org" in doi:
            doi = doi.split("doi.org/")[-1]

        # Send an HTTP GET request to fetch paper information
        response = self.client.backend.papers.GET(
            headers={"Accept": "application/vnd.orkg.paper.v2+json"},
            params={"doi": doi},
        )

        return self.client.wrap_response(response)

    def by_title(self, title: str) -> OrkgResponse:
        """
        Retrieve a paper object based on its title.
        :param title: The title of the research paper.
        :return: A list of paper objects
        """
        # Disable appending a slash to the base URL
        self.client.backend._append_slash = False

        # Send an HTTP GET request to fetch paper information
        response = self.client.backend.papers.GET(
            headers={"Accept": "application/vnd.orkg.paper.v2+json"},
            params={"title": title},
        )

        return self.client.wrap_response(response)

    @query_params("page", "size", "sort", "desc")
    def in_research_field(
        self,
        research_field_id: str,
        include_subfields: Optional[bool] = False,
        params: Optional[Dict] = None,
    ) -> OrkgResponse:
        """
        Get all papers in a research field
        :param research_field_id: the id of the research field
        :param include_subfields: True/False whether to include papers from subfields, default is False (optional)
        :param page: the page number (optional)
        :param size: number of items per page (optional)
        :param sort: key to sort on (optional)
        :param desc: true/false to sort desc (optional)
        :return: OrkgResponse object
        """
        url = self.client.backend("research-fields")(research_field_id)
        if include_subfields:
            url = url.subfields.papers
        else:
            url = url.papers
        if len(params) > 0:
            self.client.backend._append_slash = False
            response = url.GET(dict_to_url_params(params))
        else:
            self.client.backend._append_slash = True
            response = url.GET()
        return self.client.wrap_response(response)

    @staticmethod
    def _get_paper_statements_and_metadata(
        paper: Dict, params: Dict
    ) -> Tuple[Dict, Dict, Dict]:
        """
        Format paper statement values and add default values for missing data.
        Separate into dictionaries for statements and metadata.

        :param paper: dict of all info extracted from csv file
        :param params: string representation of dict (e.g. "{'name': 1}") dict with default statements, should contain CSV_PLACEHOLDER (optional)
        :return:
            - dict of paper statements (without metadata)
            - dict of paper metadata
            - dict of standard statements or empty dict if no standard statements
        """
        title = paper["paper:title"]
        authors = (
            [
                {"label": author.strip()}
                for author in str(paper["paper:authors"]).split(";")
            ]
            if "paper:authors" in paper
            and paper["paper:authors"] == paper["paper:authors"]
            else []
        )
        publication_month = paper.get("paper:publication_month", 1)
        publication_year = paper.get("paper:publication_year", 2000)
        research_field = paper.get("paper:research_field", "R11")  # "Science"
        doi = paper.get("paper:doi", "")
        url = ""
        published_in = ""
        paper_metadata = {
            "title": title,
            "authors": authors,
            "publicationMonth": publication_month,
            "publicationYear": publication_year,
            "researchField": research_field,
            "doi": doi,
            "url": url,
            "publishedIn": published_in,
        }
        standard_statements = (
            literal_eval(params["standard_statements"])
            if "standard_statements" in params
            else {}
        )
        # remove metadata from paper dict, already added above
        metadata_headers = [
            "paper:title",
            "paper:authors",
            "paper:publication_month",
            "paper:publication_year",
            "paper:doi",
            "paper:research_field",
        ]
        [paper.pop(key) if key in paper else paper for key in metadata_headers]
        return paper, paper_metadata, standard_statements

    def _add_new_paper(self, contribution_ids: List, paper: Dict) -> List:
        """
        Add new paper to ORKG instance and add contribution ID to list.

        :param contribution_ids: list of contribution ID strings
        :param paper: dict of paper statements and metadata
        :return: updated list of contribution ID strings
        """
        response = self.add(paper)
        if "id" in response.content:
            paper_id = response.content["id"]
            paper_statements = self.client.statements.get_by_subject(
                subject_id=paper_id, size=10000
            ).content
            for statement in paper_statements:
                if statement["predicate"]["id"] == "P31":
                    contribution_ids.append(statement["object"]["id"])
                    print("Added paper:", str(paper_id))
        else:
            print("Error adding paper: ", str(response.content))
        return contribution_ids

    def _add_new_contribution(
        self,
        contribution_ids: List,
        contribution_statements: Dict[str, List],
        existing_paper_id: str,
    ) -> List:
        """
        Add contribution to existing paper in ORKG instance and add contribution ID to list.

        :param contribution_ids: list of contribution ID strings
        :param contribution_statements: dict of predicate IDs and their objects
        :param existing_paper_id: paper ID string
        :return: updated list of contribution ID strings
        """
        paper_statements = self.client.statements.get_by_subject(
            subject_id=existing_paper_id, size=10000
        ).content
        contribution_amount = 0
        for paper_statement in paper_statements:
            if paper_statement["predicate"]["id"] == "P31":  # "Contribution"
                contribution_amount += 1
        contribution_id = self.client.resources.add(
            label="Contribution " + str(contribution_amount + 1),
            classes=["Contribution"],
        ).content["id"]
        self.client.statements.add(
            subject_id=existing_paper_id,
            predicate_id="P31",
            object_id=contribution_id,
        )
        self._create_statements(contribution_id, contribution_statements)
        contribution_ids.append(contribution_id)
        print(
            "Added contribution:",
            str(contribution_id),
            "to paper:",
            str(existing_paper_id),
        )
        return contribution_ids

    def _insert_research_problem(
        self, contribution_statements: Dict[str, List], research_problems: List[str]
    ) -> Dict[str, List]:
        """
        Add research problem(s) by resource ID to contribution statements dictionary.

        :param contribution_statements: dict of predicate IDs and their objects
        :param research_problems: list of (one or many) research problem(s) as strings
        :return: updated dict of contribution statements
        """
        contribution_statements["P32"] = []
        for research_problem in research_problems:
            research_problem_id = self.client.resources.find_or_add(
                label=research_problem, classes=["Problem"]
            ).content["id"]
            # P32 has research problem
            contribution_statements["P32"].append({"@id": research_problem_id})
        return contribution_statements

    def _insert_standard_statements(
        self, contribution_statements: Dict[str, List], statements_to_insert: Dict
    ) -> Dict[str, Union[str, List]]:
        """
        Create new standard statements dict with predicate IDs as keys and predicate labels as values.

        :param contribution_statements: dict of predicate IDs and their objects
        :param statements_to_insert: dict with default statements, should contain CSV_PLACEHOLDER
        :return: dict of standard statements' predicate IDs to predicate labels
        """
        for predicate in statements_to_insert:
            if isinstance(statements_to_insert[predicate], list):  # if is array
                for i in range(len(statements_to_insert[predicate])):
                    if (
                        statements_to_insert[predicate][i]["values"]
                        == "CSV_PLACEHOLDER"
                    ):
                        statements_to_insert[predicate][i][
                            "values"
                        ] = contribution_statements
            if not re.search("^P+[a-zA-Z0-9]*$", predicate):
                predicate_id = self.client.predicates.find_or_add(
                    label=predicate
                ).content["id"]
                statements_to_insert[predicate_id] = statements_to_insert[predicate]
                del statements_to_insert[predicate]
        return statements_to_insert

    def _extract_statements(self, paper: Dict) -> Tuple[List[str], Dict]:
        """
        Create a dictionary of predicate-object pairs formatted/typed for the ORKG.
        Predicates: find an existing ID matching the label, or create a new one.
        Objects: if resource, find an existing ID matching the label, or create a new one. Otherwise, add as literal.
        Also creates a list of research problem label(s).

        :param paper: dict of paper statements as extracted from CSV
        :return:
            - list of research problem label(s)
            - dict of predicate IDs and their objects
        """
        contribution_statements = {}
        research_problems = []
        for predicate in paper:
            value = paper[predicate]
            # add research problem (one or more)
            if predicate.startswith("contribution:research_problem"):
                research_problem = paper.get("predicate", "")
                if research_problem != "":
                    research_problems.append(research_problem)
                continue
            # filter out nan values
            if value != value:
                continue
            # to make columns unique, pandas appends a dot and number to duplicate columns, remove this here
            predicate = re.sub(r"\.[1-9]+$", "", predicate)
            value_is_resource = False
            # if predicate starts with resource:, insert it as resource instead of literal
            if predicate.startswith("resource:"):
                value_is_resource = True
                predicate = predicate[len("resource:") :]
            predicate_id = self.client.predicates.find_or_add(label=predicate).content[
                "id"
            ]
            if not value_is_resource:
                if predicate_id in contribution_statements:
                    contribution_statements[predicate_id].append({"text": value})
                else:
                    contribution_statements[predicate_id] = [{"text": value}]
            else:
                resource_id = self.client.resources.find_or_add(label=value).content[
                    "id"
                ]
                if predicate_id in contribution_statements:
                    contribution_statements[predicate_id].append({"@id": resource_id})
                else:
                    contribution_statements[predicate_id] = [{"@id": resource_id}]
        return research_problems, contribution_statements

    def _create_statements(self, subject_id: str, statements: Dict):
        """
        Add statements to ORKG instance.

        :param subject_id: string of subject resource ID
        :param statements: dict of predicate IDs and their objects
        """
        for predicate_id in statements:
            values = statements[predicate_id]
            for value in values:
                if "text" in value:
                    literal_id = self.client.literals.add(label=value["text"]).content[
                        "id"
                    ]
                    self.client.statements.add(
                        subject_id=subject_id,
                        predicate_id=predicate_id,
                        object_id=literal_id,
                    )
                elif "@id" in value:
                    self.client.statements.add(
                        subject_id=subject_id,
                        predicate_id=predicate_id,
                        object_id=value["@id"],
                    )
                elif "label" in value:
                    resource_id = self.client.resources.add(
                        label=value["label"]
                    ).content["id"]
                    self.client.statements.add(
                        subject_id=subject_id,
                        predicate_id=predicate_id,
                        object_id=resource_id,
                    )
                    self._create_statements(resource_id, value["values"])



from typing import Dict, Optional

from orkg.out import OrkgResponse, OrkgUnpaginatedResponse
from orkg.utils import NamespacedClient, dict_to_url_params, query_params


class StatementsClient(NamespacedClient):
    def by_id(self, id: str) -> OrkgResponse:
        self.client.backend._append_slash = True
        response = self.client.backend.statements(id).GET()
        return self.client.wrap_response(response)

    @query_params("page", "size", "sort", "desc")
    def get(self, params: Optional[Dict] = None) -> OrkgResponse:
        self.handle_sort_params(params)
        response = self.client.backend.statements.GET(params=params)
        return self.client.wrap_response(response)

    @query_params("page", "size", "sort", "desc")
    def get_unpaginated(
        self, start_page: int = 0, end_page: int = -1, params: Optional[Dict] = None
    ) -> OrkgUnpaginatedResponse:
        return self._call_pageable(
            self.get, args={}, params=params, start_page=start_page, end_page=end_page
        )

    @query_params("page", "size", "sort", "desc")
    def get_by_subject(
        self, subject_id: str, params: Optional[Dict] = None
    ) -> OrkgResponse:
        self.handle_sort_params(params)
        params.update({"subject_id": subject_id})
        response = self.client.backend.statements.GET(params=params)
        return self.client.wrap_response(response)

    @query_params("page", "size", "sort", "desc")
    def get_by_subject_unpaginated(
        self,
        subject_id: str,
        start_page: int = 0,
        end_page: int = -1,
        params: Optional[Dict] = None,
    ) -> OrkgUnpaginatedResponse:
        return self._call_pageable(
            self.get_by_subject,
            args={"subject_id": subject_id},
            params=params,
            start_page=start_page,
            end_page=end_page,
        )

    @query_params("page", "size", "sort", "desc")
    def get_by_predicate(
        self, predicate_id: str, params: Optional[Dict] = None
    ) -> OrkgResponse:
        self.handle_sort_params(params)
        params.update({"predicate_id": predicate_id})
        response = self.client.backend.statements.GET(params=params)
        return self.client.wrap_response(response)

    @query_params("page", "size", "sort", "desc")
    def get_by_predicate_unpaginated(
        self,
        predicate_id: str,
        start_page: int = 0,
        end_page: int = -1,
        params: Optional[Dict] = None,
    ) -> OrkgUnpaginatedResponse:
        return self._call_pageable(
            self.get_by_predicate,
            args={"predicate_id": predicate_id},
            params=params,
            start_page=start_page,
            end_page=end_page,
        )

    @query_params("page", "size", "sort", "desc")
    def get_by_object(
        self, object_id: str, params: Optional[Dict] = None
    ) -> OrkgResponse:
        self.handle_sort_params(params)
        params.update({"object_id": object_id})
        response = self.client.backend.statements.GET(params=params)
        return self.client.wrap_response(response)

    @query_params("page", "size", "sort", "desc")
    def get_by_object_unpaginated(
        self,
        object_id: str,
        start_page: int = 0,
        end_page: int = -1,
        params: Optional[Dict] = None,
    ) -> OrkgUnpaginatedResponse:
        return self._call_pageable(
            self.get_by_object,
            args={"object_id": object_id},
            params=params,
            start_page=start_page,
            end_page=end_page,
        )

    @query_params("page", "size", "sort", "desc")
    def get_by_object_and_predicate(
        self, object_id: str, predicate_id: str, params: Optional[Dict] = None
    ) -> OrkgResponse:
        self.handle_sort_params(params)
        params.update({"object_id": object_id, "predicate_id": predicate_id})
        response = self.client.backend.statements.GET(params=params)
        return self.client.wrap_response(response)

    @query_params("page", "size", "sort", "desc")
    def get_by_object_and_predicate_unpaginated(
        self,
        object_id: str,
        predicate_id: str,
        start_page: int = 0,
        end_page: int = -1,
        params: Optional[Dict] = None,
    ) -> OrkgUnpaginatedResponse:
        return self._call_pageable(
            self.get_by_object_and_predicate,
            args={"object_id": object_id, "predicate_id": predicate_id},
            params=params,
            start_page=start_page,
            end_page=end_page,
        )

    @query_params("page", "size", "sort", "desc")
    def get_by_subject_and_predicate(
        self, subject_id: str, predicate_id: str, params: Optional[Dict] = None
    ) -> OrkgResponse:
        self.handle_sort_params(params)
        params.update({"subject_id": subject_id, "predicate_id": predicate_id})
        response = self.client.backend.statements.GET(params=params)
        return self.client.wrap_response(response)

    @query_params("page", "size", "sort", "desc")
    def get_by_subject_and_predicate_unpaginated(
        self,
        subject_id: str,
        predicate_id: str,
        start_page: int = 0,
        end_page: int = -1,
        params: Optional[Dict] = None,
    ) -> OrkgUnpaginatedResponse:
        return self._call_pageable(
            self.get_by_subject_and_predicate,
            args={"subject_id": subject_id, "predicate_id": predicate_id},
            params=params,
            start_page=start_page,
            end_page=end_page,
        )

    @query_params("subject_id", "predicate_id", "object_id")
    def add(self, params: Optional[Dict] = None) -> OrkgResponse:
        if len(params) != 3:
            raise ValueError("all parameters must be provided")
        else:
            self.client.backend._append_slash = True
            response = self.client.backend.statements.POST(
                json=params, headers=self.auth
            )
            if response.ok:
                return self.client.expand_response(response)
        return self.client.wrap_response(response)

    @query_params("subject_id", "predicate_id", "object_id")
    def update(self, id: str, params: Optional[Dict] = None) -> OrkgResponse:
        if len(params) == 0:
            raise ValueError("at least one parameter must be provided")
        else:
            if not self.exists(id):
                raise ValueError("the provided id is not in the graph")
            self.client.backend._append_slash = True
            response = self.client.backend.statements(id).PUT(
                json=params, headers=self.auth
            )
            if response.ok:
                return self.client.expand_response(response)
        return self.client.wrap_response(response)

    @query_params("min_level", "max_level", "blacklist", "whitelist", "include_first")
    def bundle(self, thing_id: str, params: Optional[Dict] = None) -> OrkgResponse:
        if len(params) > 0:
            self.client.backend._append_slash = False
            response = self.client.backend.statements(thing_id).bundle.GET(
                dict_to_url_params(params)
            )
        else:
            self.client.backend._append_slash = True
            response = self.client.backend.statements(thing_id).bundle.GET()
        return self.client.wrap_response(response)

    def delete(self, id: str) -> OrkgResponse:
        if not self.exists(id):
            raise ValueError("the provided id is not in the graph")
        self.client.backend._append_slash = True
        response = self.client.backend.statements(id).DELETE()
        return self.client.wrap_response(response)

    def exists(self, id: str) -> bool:
        return self.by_id(id).succeeded



import json
import os
from typing import Any, List, Optional, Union

from orkg.client.harvesters.utils import process_contribution
from orkg.common import OID
from orkg.out import OrkgResponse


def harvest(
    orkg_client: Any,
    directory: str,
    research_field: Union[str, OID],
    title: str,
    doi: Optional[str] = None,
    authors: Optional[List[str]] = None,
    publication_year: Optional[int] = None,
    publication_month: Optional[int] = None,
    published_in: Optional[str] = None,
    url: Optional[str] = None,
    extraction_method: Optional[str] = None,
    slow_mode: Optional[bool] = False,
    **kwargs,
) -> OrkgResponse:
    # Check if the directory exists
    if not os.path.isdir(directory):
        raise ValueError(f"The directory {directory} does not exist.")

    # Check if the directory contains any json files
    if not any(file.endswith(".json") for file in os.listdir(directory)):
        raise ValueError(f"The directory {directory} does not contain any json files.")

    # Check if research field is valid
    if isinstance(research_field, str):
        rf_response = orkg_client.resources.get(q=research_field, exact=True, size=1)
        if not rf_response.succeeded or len(rf_response.content) == 0:
            raise ValueError(
                f"Unable to find the ORKG research field with the given string value {research_field}"
            )
        research_field = OID(rf_response.content[0]["id"])

    # Prepare the paper object
    paper_json = {"title": title, "researchField": f"{research_field.id}"}
    if doi is not None:
        paper_json["doi"] = doi
    if authors is not None:
        paper_json["authors"] = [
            {"label": author} if isinstance(author, str) else {"id": str(author)}
            for author in authors
        ]
    if publication_year is not None:
        paper_json["publicationYear"] = str(publication_year)
    if publication_month is not None:
        paper_json["publicationMonth"] = str(publication_month)
    if published_in is not None:
        paper_json["publishedIn"] = published_in
    if url is not None:
        paper_json["url"] = url
    if extraction_method is not None:
        paper_json["extractionMethod"] = extraction_method
    if "contributions" not in paper_json:
        paper_json["contributions"] = []
    for key, value in kwargs.items():
        paper_json[key] = value

    # Read all json files in the directory
    contribution_content = []
    for file in os.listdir(directory):
        if not file.endswith(".json"):
            continue
        with open(os.path.join(directory, file)) as f:
            contribution_content.append(json.load(f))

    # Process the contribution content
    for contribution_json in contribution_content:
        orkg_contribution_json = {}
        global_ids = {}
        context = contribution_json.get("@context", {})
        process_contribution(
            contribution_json, orkg_contribution_json, global_ids, context
        )
        # replace the key "label" with "name"
        orkg_contribution_json["name"] = orkg_contribution_json.pop("label")
        paper_json["contributions"].append(orkg_contribution_json)

    # Remove @temp from all the contribution objects (NamedObject doesn't have a @temp field anymore)
    for contribution in paper_json["contributions"]:
        if "@temp" in contribution:
            del contribution["@temp"]

    # Now that we have everything, let's finalize the paper object and add it to the graph
    paper_json = {"paper": paper_json}
    if not slow_mode:
        return orkg_client.papers.add(paper_json)
    else:
        contributions = paper_json["paper"]["contributions"]
        paper_json["paper"]["contributions"] = [contributions[0]]
        paper_response = orkg_client.papers.add(paper_json)
        for contribution in contributions[1:]:
            contribution_response = orkg_client.objects.add({"resource": contribution})
            orkg_client.statements.add(
                subject_id=paper_response.content["id"],
                predicate_id="P31",
                object_id=contribution_response.content["id"],
            )
        return paper_response



from typing import Dict, Optional

from orkg.out import OrkgResponse
from orkg.utils import NamespacedClient, dict_to_url_params, query_params


class ResearchFieldsClient(NamespacedClient):
    @query_params("page", "size", "sort", "desc")
    def with_benchmarks(self, params: Optional[Dict] = None) -> OrkgResponse:
        """
        Get all the research fields with benchmarks
        :param page: the page number (optional)
        :param size: number of items per page (optional)
        :param sort: key to sort on (optional)
        :param desc: true/false to sort desc (optional)
        :return: OrkgResponse object
        """
        if len(params) > 0:
            self.client.backend._append_slash = False
            response = self.client.backend("research-fields").benchmarks.GET(
                dict_to_url_params(params)
            )
        else:
            self.client.backend._append_slash = True
            response = self.client.backend("research-fields").benchmarks.GET()
        return self.client.wrap_response(response)

    @query_params("page", "size", "sort", "desc")
    def get_subfields(
        self, field_id: str, params: Optional[Dict] = None
    ) -> OrkgResponse:
        """
        Get the subfield(s) immediately below the given research field
        :param field_id: id of the research field
        :param page: the page number (optional)
        :param size: number of items per page (optional)
        :param sort: key to sort on (optional)
        :param desc: true/false to sort desc (optional)
        :return: OrkgResponse object
        """
        if len(params) > 0:
            self.client.backend._append_slash = False
            response = self.client.backend("research-fields")(field_id).children.GET(
                dict_to_url_params(params)
            )
        else:
            self.client.backend._append_slash = True
            response = self.client.backend("research-fields")(field_id).children.GET()
        return self.client.wrap_response(response)

    @query_params("page", "size", "sort", "desc")
    def get_superfields(
        self, field_id: str, params: Optional[Dict] = None
    ) -> OrkgResponse:
        """
        Get the superfield(s) immediately above the given research field
        :param field_id: id of the research field
        :param page: the page number (optional)
        :param size: number of items per page (optional)
        :param sort: key to sort on (optional)
        :param desc: true/false to sort desc (optional)
        :return: OrkgResponse object
        """
        if len(params) > 0:
            self.client.backend._append_slash = False
            response = self.client.backend("research-fields")(field_id).parents.GET(
                dict_to_url_params(params)
            )
        else:
            self.client.backend._append_slash = True
            response = self.client.backend("research-fields")(field_id).parents.GET()
        return self.client.wrap_response(response)

    def get_root(self, field_id: str) -> OrkgResponse:
        """
        Get the root research field of the given subfield
        :param field_id: id of the subfield
        :return: OrkgResponse object
        """
        self.client.backend._append_slash = True
        response = self.client.backend("research-fields")(field_id).roots.GET()
        return self.client.wrap_response(response)

    @query_params("page", "size", "sort", "desc")
    def get_all_roots(self, params: Optional[Dict] = None) -> OrkgResponse:
        """
        Get a list of all root research fields
        :param page: the page number (optional)
        :param size: number of items per page (optional)
        :param sort: key to sort on (optional)
        :param desc: true/false to sort desc (optional)
        :return: OrkgResponse object
        """
        if len(params) > 0:
            self.client.backend._append_slash = False
            response = self.client.backend("research-fields").roots.GET(
                dict_to_url_params(params)
            )
        else:
            self.client.backend._append_slash = True
            response = self.client.backend("research-fields").roots.GET()
        return self.client.wrap_response(response)

    @query_params("page", "size", "sort", "desc")
    def get_hierarchy(
        self, field_id: str, params: Optional[Dict] = None
    ) -> OrkgResponse:
        """
        Get a list of all research fields in the hierarchy above and including the given subfield
        :param field_id: id of the subfield
        :param page: the page number (optional)
        :param size: number of items per page (optional)
        :param sort: key to sort on (optional)
        :param desc: true/false to sort desc (optional)
        :return: OrkgResponse object
        """
        if len(params) > 0:
            self.client.backend._append_slash = False
            response = self.client.backend("research-fields")(field_id).hierarchy.GET(
                dict_to_url_params(params)
            )
        else:
            self.client.backend._append_slash = True
            response = self.client.backend("research-fields")(field_id).hierarchy.GET()
        return self.client.wrap_response(response)

    @query_params("include_subfields")
    def get_stats(self, field_id: str, params: Optional[Dict] = None) -> OrkgResponse:
        """
        Get the number of papers within the given research field, optionally including all subfields
        :param field_id: id of the subfield
        :param include_subfields: True/False whether to include subfields in stats, default is False (optional)
        :return: OrkgResponse object
        """
        params_to_send = {
            "research_field": field_id,
            "include_subfields": params.get("include_subfields", False),
        }
        self.client.backend._append_slash = False
        response = self.client.backend.statistics("content-types")("paper-count").GET(
            dict_to_url_params(params_to_send)
        )
        return self.client.wrap_response(response)

    @query_params("page", "size", "sort", "desc")
    def for_research_problem_id(
        self, research_problem_id: str, params: Optional[Dict] = None
    ) -> OrkgResponse:
        """
        Get all research fields associated with a research problem
        :param research_problem_id: id of the research problem
        :param page: the page number (optional)
        :param size: number of items per page (optional)
        :param sort: key to sort on (optional)
        :param desc: true/false to sort desc (optional)
        :return: OrkgResponse object
        """
        if params is None:
            params = {}
        params["research_problem"] = research_problem_id

        self.client.backend._append_slash = False
        response = self.client.backend("research-fields").GET(
            dict_to_url_params(params)
        )
        return self.client.wrap_response(response)



import json
from typing import Dict

from orkg.client.templates.statements import ObjectStatement
from orkg.convert import JsonLDConverter
from orkg.out import OrkgResponse


class TemplateInstance(object):
    """
    A class that takes a dictionary in the constructor with methods
    to serialize the dictionary to a file, pretty print it, and send it to the KG.
    """

    def __init__(self, template_dict, orkg_client) -> None:
        self.template_dict = template_dict
        self.client = orkg_client
        self.preprocess_dict()

    def preprocess_dict(self):
        if isinstance(self.template_dict, str):
            self.template_dict = (
                self.template_dict.strip("'<>() ")
                .replace("'", '"')
                .replace("{{", "{")
                .replace("}}", "}")
                .replace('""', '"')
            )
            self.template_dict = json.loads(self.template_dict)

    def serialize_to_file(self, file_path: str, format: str = "orkg") -> None:
        """
        Serializes the template to a file.
        :param format: the format of the serialization (default: "orkg", possible: "json-ld")
        :param file_path: the file path to save the template to
        """
        with open(file_path, "w") as f:
            if format.lower().strip() == "json-ld":
                json.dump(
                    JsonLDConverter(self.client.host).convert_2_jsonld(
                        self.template_dict
                    ),
                    f,
                    indent=4,
                )
            else:
                json.dump(self.template_dict, f, indent=4)

    def pretty_print(self, format: str = "orkg") -> None:
        """
        Pretty prints the template to the console.
        :param format: the format of the printed text (default: "orkg", possible: "json-ld")
        """
        if format.lower().strip() == "json-ld":
            print(
                json.dumps(
                    JsonLDConverter(self.client.host).convert_2_jsonld(
                        self.template_dict
                    ),
                    indent=4,
                )
            )
        else:
            print(json.dumps(self.template_dict, indent=4))

    def save(self) -> OrkgResponse:
        """
        Saves the template to the server.
        :return: The OrkgResponse from the server
        """
        return self.client.objects.add(params=self.template_dict)

    @staticmethod
    def parse_dict(dictionary: Dict, obj: ObjectStatement) -> None:
        if "resource" in dictionary:
            # This is a complete resource and not a nested one
            dictionary = dictionary["resource"]
            obj.set_resource(dictionary["name"], dictionary["classes"])
            if "values" in dictionary:
                dictionary = dictionary["values"]
        for key, values in dictionary.items():
            for value in values:
                sub_obj = ObjectStatement(key, obj.template_id)
                obj.add_value(sub_obj)
                if "text" in value:
                    datatype = value.get("datatype", None)
                    sub_obj.set_literal(f"\"{value['text']}\"", datatype)
                elif "label" in value:
                    classes = value.get("classes", None)
                    sub_obj.set_resource(f"\"{value['label']}\"", classes)
                    if "values" in value:
                        TemplateInstance.parse_dict(value["values"], sub_obj)
